"""Packages containing sleep events detection

"""
from .fish import FISH, extract_data_error, extract_pac_data, extract_event_data
from .surf import erp_analysis, insert_virtual_markers, marker_to_annot
from .seasnakes import seasnakes, swordfish
from .whales import whales